#! /usr/bin/env python

## Copyright 2000-2001 by LivingLogic AG, Bayreuth, Germany.
## Copyright 2000-2001 by Walter Drwald
##
## All Rights Reserved
##
## Permission to use, copy, modify, and distribute this software and its documentation
## for any purpose and without fee is hereby granted, provided that the above copyright
## notice appears in all copies and that both that copyright notice and this permission
## notice appear in supporting documentation, and that the name of LivingLogic AG or
## the author not be used in advertising or publicity pertaining to distribution of the
## software without specific, written prior permission.
##
## LIVINGLOGIC AG AND THE AUTHOR DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
## INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
## LIVING LOGIC AG OR THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
## DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
## IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
## IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

r"""
<par>This module simplifies running Python stuff as cron jobs.</par>

<par>There will be no more than one sisyphus job of a certain name
running at every given time. When the job is already running
and a second is started, the second one will quit immediately
if the first one hasn't exceeded its maximum allowed lifetime
yet. If it has exceeded the allowed lifetime the first job
will be killed and the second will start running.</par>

<par>In addition to that, logging features are provided.</par>

<par>To use this module, you must derive your own
class from <code>sisyphus.Job</code> and implement
the <code>execute</code> method.</par>

<par>The job announces its presence (and its process id) in a file that
is stored in the <code>~/run</code> directory. Logs will be created in the
<code>~/log</code> directory (This can be changes by deriving new subclasses).</par>

<par>There are three log files:
<ol>
<li><code>~/log/<var>jobname</var>_progress.log</code><br/>
The progress of one job execution is logged here.
This log file will be truncated at the start of every execution,
so you can be rather verbose.
Use the <code>logProgress</code> method for writing to this log file.</li>
<li><code>~/log/<var>jobname</var>_loop.log</code><br/>
One log line may be appended to the log after every job execution.
Use the <code>logLoop</code> method for this.</li>
<li><code>~/log/<var>jobname</var>_error.log</code><br/>
Here exceptions happening during the execution of a job
will be logged. This is done via the <code>logError</code> method
with can be used for reporting other exception conditions apart
from exceptions.</li>
</ol>
</par>

<par>To execute a job, use the function <code>sisyphus.execute</code>.
Arguments are job objects, that will be executed sequentially.</par>

<h2>Example</h2>

<par>The following example illustrates the use of this module:
<example title="Complete example showing the use of sisyphus">
<programlisting>
#!/usr/bin/env python

import os
import urllib
import sisyphus

class Fetch(sisyphus.Job):
	"savely fetches an HTML file and saves it to a local file."
	def __init__(self):
		sisyphus.Job.__init__(self, 180, name="Fetch")
		self.url = "http://www.python.org/"
		self.tmpname = "Fetch_Tmp_%d.html" % os.getpid()
		self.officialname = "Python.html"

	def execute(self):
		self.logProgress("fetching data from %r" % self.url)
		data = urllib.urlopen(self.url).read()
		datasize = len(data)
		self.logProgress("writing file %r (%d bytes)" %
			(self.tmpname, datasize))
		open(self.tmpname, "wb").write(data)
		self.logProgress("renaming file %r to %r" %
			(self.tmpname, self.officialname))
		os.rename(self.tmpname, self.officialname)
		self.logLoop("cached %r as %r (%d bytes)" %
			(self.url, self.officialname, datasize))

if __name__=="__main__":
	sisyphus.execute(Fetch())
</programlisting>
</example>
</par>
"""

__version__ = tuple(map(int,"$Revision: 1.46 $"[11:-2].split(".")))
# $Source: /data/cvsroot/LivingLogic/sisyphus/sisyphus.py,v $

import sys
import os
import os.path
import exceptions
import stat
import time
import traceback

import fileutils

class Job:
	"""
	A Job object executes a task once.
	
	The job announces its presence (and its process id) in a file that
	is stored in the ~/run directory. Logs will be created in the
	~/log directory (This can be changes by deriving new subclasses).

	There are three log files:
	1. ~/log/<jobname>_progress.log
	   The progress of one job execution is logged here.
	   This log file will be truncated at the start of every execution.
	   Use the logProgress() method for writing to this log file.
	2. ~/log/<jobname>_loop.log
	   One log line may be appended to the log after every job
	   execution.
	   Use the logLoop() method for this
	3. ~/log/<jobname>_error.log
	   Here exceptions happening during the execution of a job
	   will be logged.
	   This is done via the logError method with can be used
	   for reporting additional errors.

	To use this class, derive your own class from it and overwrite
	the execute method.
	"""

	pidFilenameTemplate = "~/run/%s.pid"
	loopFilenameTemplate = "~/log/%s_loop.log"
	errorFilenameTemplate = "~/log/%s_error.log"
	progressFilenameTemplate = "~/log/%s_progress.log"

	def __init__(self, maxRuntime=0, name=None, raiseErrors=0):
		"""
		__init__(maxRuntime=0, name=None) -- construct a Job object.
		
		maxRuntime -- the maximum allowed runtime in seconds for this job.
		name -- the name to be used for the log files. If None, the name 
		  of the class will be used.
		raiseErrors -- should exceptions that occur during the excution of the
		  job be raised (which results in a output to the terminal, or an email
		  from the cron daemon)
		"""
		self.startTime = time.time()
		self.maxRuntime = maxRuntime
		self.name = name or self.__class__.__name__
		self.raiseErrors = raiseErrors
		self.pidFileWritten = 0
		self.pidFilename = self.pidFilenameTemplate % self.name
		self.loopLogfile = fileutils.LogFile(self.loopFilenameTemplate % self.name)
		self.errorLogfile = fileutils.LogFile(self.errorFilenameTemplate % self.name)
		self.progressLogfile = fileutils.LogFile(self.progressFilenameTemplate % self.name, mode="w")

	def __writePid(self):
		"""
		creates the file containing the pid of the current process
		"""
		file = fileutils.Filename(self.pidFilename).open(mode="w")
		file.write(str(os.getpid()))
		file.close()
		self.pidFileWritten = 1

	def __killPid(self):
		"""
		deletes the pid file
		"""
		if self.pidFileWritten:
			fileutils.remove(self.pidFilename)
			self.pidFileWritten = 0

	def logLoop(self, *texts):
		"""
		log the message texts to the loop and progress log.
		(The call to logLoop() should be the last statement
		in the execute method.)
		"""
		self.loopLogfile.write(*texts)
		self.progressLogfile.write(*texts)

	def logProgress(self, *texts):
		"""
		log the message texts to the progress log
		"""
		self.progressLogfile.write(*texts)

	def logError(self, error):
		"""
		log the error to the error log.
		error may be a string or an exception object.
		"""
		if isinstance(error, exceptions.Exception):
			tb = "\n" + "".join(traceback.format_tb(sys.exc_info()[-1]))
			name = error.__class__.__name__
			tb += "%s: %s" % (name, error)
			self.errorLogfile.write(tb)
		else:
			self.errorLogfile.write(error)

	def execute(self):
		"""
		this method executes the job once.
		At the end of the job you should write one line to the loop log.
		Overwrite in subclasses.
		"""
		self.logLoop("done")

	def handleExecution(self):
		"""
		handles executing the job including handling of duplicate or hanging jobs.
		This is the method to be called from the outside world.
		"""
		try: # is there a pid file from a running job?
			file = open(os.path.expanduser(self.pidFilename), "r")
		except IOError, ex: # no pid file => the job has been finished without problems
			if ex[0] == 2: # file not found
				self.__writePid()
				self.logProgress("no previous job running; here we go")
			else:
				raise
		else:
			lastModified = os.fstat(file.fileno())[stat.ST_MTIME]
			pid = int(file.read())
			file.close()
			if self.startTime-lastModified > self.maxRuntime: # the job is to old, so it probably hangs => kill it
				try:
					os.kill(pid, 9)
				except OSError: # there was no process
					pass
				self.__writePid()
				self.logError("killed previous job running with pid %d (ran %.2f seconds; %d allowed); here we go" % (pid, self.startTime-lastModified, self.maxRuntime))
			else:
				self.logError("Job still running (for %.2f seconds; %d allowed; started on %s) with pid %d (according to %s)" % (self.startTime-lastModified, self.maxRuntime, time.strftime("%d/%b/%Y %H:%M:%S", time.localtime(lastModified)), pid, self.pidFilename))
				return

		try:
			self.execute()
		except Exception, ex:
			self.logError(ex) # log the error
			self.logLoop("failed with %s" % ex.__class__.__name__) # log the error to the lop loog too, because the job probably didn't have a chance to do it.
			if self.raiseErrors or isinstance(ex, KeyboardInterrupt): # Really exit
				self.__killPid()
				raise
		self.__killPid() # finished => remove the pid file

def execute(*jobs):
	"""
	sequentially execute several jobs.
	"""
	for job in jobs:
		job.handleExecution()

